
struct complex {
    // Set imag_zero to true and these numbers will
    // (inefficiently) behave like real numbers.
    static constexpr bool imag_zero {false};
private:
    double x, y;
    static constexpr double tol {1.0e-10};
public:
    complex (complex const &) = default;
    complex (complex &&) = default;
    complex & operator= (complex const &) = default;
    complex & operator= (complex &&) = default;
    ~complex () = default;

    explicit complex (double const a = double{}, double const b = double{}) :
        x {a}, y {b}
    {
        if (! imag_zero) return;
        if (y != double{}) throw fail {"complex ... imag_zero violation"};
    }

    static complex random ()
        { return complex { {rd()}, {imag_zero ? double{} : rd()} }; }

    complex operator+ () const
        { return complex {+x, +y}; }
    complex operator~ () const
        { return complex {+x, -y}; }
    complex operator- () const
        { return complex {-x, -y}; }

    complex operator+ (complex const & o) const
        { return complex {x + o.x, y + o.y}; }
    complex & operator+= (complex const & o)
        { x += o.x; y += o.y; return *this; }

    complex operator- (complex const & o) const
        { return complex {x - o.x, y - o.y}; }
    complex & operator-= (complex const & o)
        { x -= o.x; y -= o.y; return *this; }

    complex operator* (complex const & o) const {
        double const a {x * o.x - y * o.y};
        double const b {x * o.y + y * o.x};
        return complex {a, b};
    }
    complex & operator*= (complex const & o) {
        complex const temp {(*this) * o};
        *this = temp;
        return *this;
    }

    complex operator* (double const & o) const {
        return complex {x * o, y * o};
    }

    complex operator/ (double const & o) const {
        return complex {x/o, y/o};
    }
    complex operator/ (complex const & o) const {
        double const osm {o.squ_mag()};
        return complex {operator* (~o) / osm};
    }

    bool operator== (complex const & o) const {
        if (x < (o.x - tol)) return false;
        if (x > (o.x + tol)) return false;
        if (y < (o.y - tol)) return false;
        if (y > (o.y + tol)) return false;
        return true;
    }

    bool operator!= (complex const & o) const {
        if (x < (o.x - tol)) return true;
        if (x > (o.x + tol)) return true;
        if (y < (o.y - tol)) return true;
        if (y > (o.y + tol)) return true;
        return false;
    }

    double real () const { return x; }
    double imag () const { return y; }
    
    double squ_mag () const { return x * x + y * y; }
    double mag () const { return sqrt (squ_mag()); }
};

ostream & operator<< (ostream & o, complex const & c) {
    cout << showpos;
    if (complex::imag_zero)
        o << c.real();
    else 
        o << "(" << c.real() << ", " << c.imag() << ")";
    return o;
}