
struct eight_dim {
    // Set four_only to true and the vectors will
    // (inefficiently) behave like they are four-dimensional.
    // This gives you the 3-in-4 cross product.
    static constexpr bool four_only {false};
    // For the star product, four_only MUST be false.
private:
	vector <complex> guts;
public:
    eight_dim (eight_dim const &) = default;
    eight_dim (eight_dim &&) = default;
    eight_dim & operator= (eight_dim const &) = default;
    eight_dim & operator= (eight_dim &&) = default;
    ~eight_dim () = default;

	eight_dim () : guts (8) { }

	eight_dim (
		complex const a,
        complex const b,
        complex const c,
        complex const d,
		complex const e = complex{},
        complex const f = complex{},
        complex const g = complex{},
        complex const h = complex{}
	) : guts {a,b,c,d,e,f,g,h} {
        if (! four_only) return;
        if (e != complex{}) throw fail {"eight_dim ... four_only violation"};
        if (f != complex{}) throw fail {"eight_dim ... four_only violation"};
        if (g != complex{}) throw fail {"eight_dim ... four_only violation"};
        if (h != complex{}) throw fail {"eight_dim ... four_only violation"};
	}

    static eight_dim random () {
        return eight_dim {
            complex::random(),
            complex::random(),
            complex::random(),
            complex::random(),
            four_only ? complex {}: complex::random(),
            four_only ? complex {}: complex::random(),
            four_only ? complex {}: complex::random(),
            four_only ? complex {}: complex::random()
        };
    }

    eight_dim operator+ () const {
        return eight_dim {
            +guts.at(0), +guts.at(1), +guts.at(2), +guts.at(3),
            +guts.at(4), +guts.at(5), +guts.at(6), +guts.at(7)
        };
    }

    eight_dim operator- () const {
        return eight_dim {
            -guts.at(0), -guts.at(1), -guts.at(2), -guts.at(3),
            -guts.at(4), -guts.at(5), -guts.at(6), -guts.at(7)
        };
    }

    // observe difference between operator~ and operator!

        // vector-style conjugation
        eight_dim operator~ () const {
            return eight_dim {
                ~guts.at(0), ~guts.at(1), ~guts.at(2), ~guts.at(3),
                ~guts.at(4), ~guts.at(5), ~guts.at(6), ~guts.at(7)
            };
        }

        // octonion-style conjugation
        // real part in position 3, to coordinate with chunk::octonion_table
        eight_dim operator! () const {
            return eight_dim {
                -guts.at(0), -guts.at(1), -guts.at(2), +guts.at(3),
                -guts.at(4), -guts.at(5), -guts.at(6), -guts.at(7)
            };
        }

	eight_dim operator+ (eight_dim const & o) const {
		return eight_dim {
			guts.at(0) + o.guts.at(0), guts.at(1) + o.guts.at(1),
			guts.at(2) + o.guts.at(2), guts.at(3) + o.guts.at(3),
			guts.at(4) + o.guts.at(4), guts.at(5) + o.guts.at(5),
			guts.at(6) + o.guts.at(6), guts.at(7) + o.guts.at(7)
		};
	}
    eight_dim & operator+= (eight_dim const & o) {
        for (int n {0}; n < guts.size(); ++n)
            guts.at(n) += o.guts.at(n);
        return *this;
    }

	eight_dim operator- (eight_dim const & o) const {
		return eight_dim {
			guts.at(0) - o.guts.at(0), guts.at(1) - o.guts.at(1),
			guts.at(2) - o.guts.at(2), guts.at(3) - o.guts.at(3),
			guts.at(4) - o.guts.at(4), guts.at(5) - o.guts.at(5),
			guts.at(6) - o.guts.at(6), guts.at(7) - o.guts.at(7)
		};
	}

	eight_dim operator* (complex const & o) const {
		return eight_dim {
			guts.at(0) * o, guts.at(1) * o,
			guts.at(2) * o, guts.at(3) * o,
			guts.at(4) * o, guts.at(5) * o,
			guts.at(6) * o, guts.at(7) * o
		};
	}

	eight_dim operator* (double const o) const {
		return eight_dim {
			guts.at(0) * o, guts.at(1) * o,
			guts.at(2) * o, guts.at(3) * o,
			guts.at(4) * o, guts.at(5) * o,
			guts.at(6) * o, guts.at(7) * o
		};
	}

	eight_dim operator/ (complex const & o) const {
		return eight_dim {
			guts.at(0) / o, guts.at(1) / o,
			guts.at(2) / o, guts.at(3) / o,
			guts.at(4) / o, guts.at(5) / o,
			guts.at(6) / o, guts.at(7) / o
		};
	}

	eight_dim operator/ (double const o) const {
		return eight_dim {
			guts.at(0) / o, guts.at(1) / o,
			guts.at(2) / o, guts.at(3) / o,
			guts.at(4) / o, guts.at(5) / o,
			guts.at(6) / o, guts.at(7) / o
		};
	}

    eight_dim slice (int const w) const {
        static int const table[14][4] {
            {0,1,2,3}, {4,5,6,7},
            {0,1,4,5}, {2,3,6,7},
            {0,1,6,7}, {2,3,4,5},
            {0,2,4,6}, {1,3,5,7},
            {0,2,5,7}, {1,3,4,6},
            {0,3,4,7}, {1,2,5,6},
            {0,3,5,6}, {1,2,4,7}
        };

        if (w <  0) throw fail {"eight_dim::slice ... w <  0"};
        if (w > 13) throw fail {"eight_dim::slice ... w > 13"};
        int const a {table[w][0]};
        int const b {table[w][1]};
        int const c {table[w][2]};
        int const d {table[w][3]};

        eight_dim result {};
        result.guts.at(a) = this->guts.at(a);
        result.guts.at(b) = this->guts.at(b);
        result.guts.at(c) = this->guts.at(c);
        result.guts.at(d) = this->guts.at(d);
        return result;
    }

	complex & at (int const n)       { return guts.at (n); }
	complex   at (int const n) const { return guts.at (n); }

    double squ_mag () const {
        double result {};
        for (auto n : guts) result += n.squ_mag();
        return result;
    }
    double mag () const { return sqrt (squ_mag()); }

    eight_dim unitize () const {
        return this->operator/ (this->mag());
    }

    chunk as_unit () const {
        int sgn { 0};
        int vec {-1};

        for (int n {0}; n < 8; ++n) {
            complex const & gan {guts.at(n)};

            if (gan == complex {0.0})
                ;
            else if (vec != -1)
                throw fail {"eight_dim::parse ... duplicate"};
            else if (gan == complex {+1.0})
                { sgn = +1; vec = n; }
            else if (gan == complex {-1.0})
                { sgn = -1; vec = n; }
            else
                throw fail {"eight_dim::parse ... mystery value"};
        }

        if (vec == -1) return chunk::ZZ;
        return chunk {sgn, vec};
    }
};

ostream & operator<< (ostream & o, eight_dim const & ed) {
	o << "[ ";
    if (eight_dim::four_only)
        for (int n (0); n < 4; ++n) o << ed.at (n) << " ";
    else
        for (int n (0); n < 8; ++n) o << ed.at (n) << " ";
	o << "]";
	return o;
}
