
// conjugation intentionally omitted from complex dot product
complex dot (eight_dim const & p, eight_dim const & q) {
	complex r {0.0, 0.0};
	for (int a (0); a < 8; ++a) r += p.at(a) * q.at(a);
	return r;
}

complex gram ( // determinant
	eight_dim const & p, eight_dim const & q, eight_dim const & r
) {
	complex const pp (dot (p, p));
	complex const pq (dot (p, q));
	complex const pr (dot (p, r));
	complex const qp (dot (q, p));
	complex const qq (dot (q, q));
	complex const qr (dot (q, r));
	complex const rp (dot (r, p));
	complex const rq (dot (r, q));
	complex const rr (dot (r, r));

	complex const a (pp * qq * rr);
	complex const b (pq * qr * rp);
	complex const c (pr * rq * qp);
	complex const d (pp * rq * qr);
	complex const e (qq * pr * rp);
	complex const f (rr * qp * pq);

	return a + b + c - d - e - f;
}

eight_dim octonion_mult (
    eight_dim const & p, eight_dim const & q
) {
    // assume real component in position 3
    // assume imaginary components in other positions

    eight_dim r;

	for (int a {0}; a < 8; ++a) {
	for (int b {0}; b < 8; ++b) {
		chunk const & u {chunk::octonion(a, b)};
		if (u.sgn == +1)
			r.at(u.vec) += p.at(a) * q.at(b);
		else if (u.sgn == -1) 
			r.at(u.vec) -= p.at(a) * q.at(b);
	} }		

    return r;
}

///////////////////

eight_dim baseline ( // a cross product
	eight_dim const & p, eight_dim const & q, eight_dim const & r
) {
	eight_dim s;

	for (int a {0}; a < 8; ++a) {
	for (int b {0}; b < 8; ++b) {
	for (int c {0}; c < 8; ++c) {
		chunk const & u (chunk::baseline(a,b,c));
		if (u.sgn == +1)
			s.at(u.vec) += p.at(a) * q.at(b) * r.at(c);
		else if (u.sgn == -1) 
			s.at(u.vec) -= p.at(a) * q.at(b) * r.at(c);
	} } }		

	return s;
}

eight_dim lee_leung ( // a cross product
    eight_dim const & p, eight_dim const & q, eight_dim const & r
) {
    eight_dim const X {octonion_mult (p, octonion_mult (!q, r))};
    eight_dim const Y {octonion_mult (r, octonion_mult (!q, p))};
    return (X - Y) * 0.5;
}

eight_dim gray_1 ( // a cross product
    eight_dim const & p, eight_dim const & q, eight_dim const & r
) {
    eight_dim const W {octonion_mult (octonion_mult (p, !q), r)};
    eight_dim const X {q * dot (p, r)};
    eight_dim const Y {p * dot (q, r)};
    eight_dim const Z {r * dot (p, q)};
    return - W - X + Y + Z;
}

eight_dim gray_2 ( // a cross product
    eight_dim const & p, eight_dim const & q, eight_dim const & r
) {
    eight_dim const W {octonion_mult (p, octonion_mult (!q, r))};
    eight_dim const X {q * dot (p, r)};
    eight_dim const Y {p * dot (q, r)};
    eight_dim const Z {r * dot (p, q)};
    return - W - X + Y + Z;
}

eight_dim star ( // not a cross product, but interesting
	eight_dim const & p, eight_dim const & q, eight_dim const & r
) {
    if (eight_dim::four_only == true)
        throw fail {"star ... cannot have star product in four dimensions"};

	eight_dim s;

	for (int a {0}; a < 8; ++a) {
	for (int b {0}; b < 8; ++b) {
	for (int c {0}; c < 8; ++c) {
		chunk const & u (chunk::star(a,b,c));
		if (u.sgn == +1)
			s.at(u.vec) += p.at(a) * q.at(b) * r.at(c);
		else if (u.sgn == -1) 
			s.at(u.vec) -= p.at(a) * q.at(b) * r.at(c);
	} } }		

	return s;
}

///////////////////

eight_dim power_sinisterior (
    eight_dim (* oper) (
        eight_dim const &, eight_dim const &, eight_dim const &
    ),
    eight_dim const & p, eight_dim const & q, eight_dim const & r,
    int const x
) {
    if (x < 0) throw fail {"cross_power_sinisterior ... negative exponent"};
    eight_dim s {r};
    for (int n {0}; n < x; ++n) s = oper (p, q, s);
    return s;
}

eight_dim power_exterior (
    eight_dim (* oper) (
        eight_dim const &, eight_dim const &, eight_dim const &
    ),
    eight_dim const & p, eight_dim const & q, eight_dim const & r,
    int const x
) {
    if (x < 0) throw fail {"cross_power_exterior ... negative exponent"};
    eight_dim s {q};
    for (int n {0}; n < x; ++n) s = oper (p, s, r);
    return s;
}

eight_dim power_dexterior (
    eight_dim (* oper) (
        eight_dim const &, eight_dim const &, eight_dim const &
    ),
    eight_dim const & p, eight_dim const & q, eight_dim const & r,
    int const x
) {
    if (x < 0) throw fail {"cross_power_dexterior ... negative exponent"};
    eight_dim s {p};
    for (int n {0}; n < x; ++n) s = oper (s, q, r);
    return s;
}


